@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # Copyright (C) 2016, the Pyzo development team
 #
 # Pyzo is distributed under the terms of the 2-Clause BSD License.
@@ -117,8 +116,7 @@ class KernelInfo(ssdf.Struct):
                 s = ssdf.loads(info)
             else:
                 raise ValueError(
-                    "Kernel info should be a string or ssdf struct, not %s"
-                    % str(type(info))
+                    "Kernel info should be a string or ssdf struct, not {}".format(type(info))
                 )
             # Inject values
             for key in s:
@@ -169,7 +167,7 @@ def getEnvFromKernelInfo(info):
     pythonPaths = [p.strip() for p in pythonPath.split("\n") if p]
     # Recombine using the OS's path separator
     pythonPath = os.pathsep.join(pythonPaths)
-    # Add entry to Pythopath, so that we can import yoton
+    # Add entry to Pythonpath, so that we can import yoton
     # Note: an empty entry might cause trouble if the start-directory is
     # somehow overriden (see issue 128).
     pythonPath = pyzo.pyzoDir + os.pathsep + pythonPath
@@ -181,6 +179,7 @@ def getEnvFromKernelInfo(info):
     #
     env.pop("TK_LIBRARY", "")
     env.pop("TCL_LIBRARY", "")
+    env.pop("QT_MAC_WANTS_LAYER")
     env["PYTHONPATH"] = pythonPath
     # Jython does not use PYTHONPATH but JYTHONPATH
     env["JYTHONPATH"] = pyzo.pyzoDir + os.pathsep + os.environ.get("JYTHONPATH", "")
@@ -244,7 +243,7 @@ class KernelBroker:
     This class functions as a broker between a kernel process and zero or
     more IDE's (clients).
 
-    This class has a single context assosiated with it, that lives as long
+    This class has a single context associated with it, that lives as long
     as this object. It is used to connect to a kernel process and to
     0 or more IDE's (clients). The kernel process can be "restarted", meaning
     that it is terminated and a new process started.
@@ -310,11 +309,10 @@ class KernelBroker:
         self._reqp_introspect = yoton.ReqChannel(ct, "reqp-introspect")
 
     def _reset(self, destroy=False):
-        """_reset(destroy=False)
+        """Reset state.
 
-        Reset state. if destroy, does a full clean up, closing the context
+        If destroy, does a full clean up, closing the context
         and removing itself from the KernelManager's list.
-
         """
 
         # Close connection (it might be in a wait state if the process
@@ -355,29 +353,18 @@ class KernelBroker:
             self._reqp_introspect = None
 
     def startKernelIfConnected(self, timeout=10.0):
-        """startKernelIfConnected(timout=10.0)
-
-        Start the kernel as soon as there is a connection.
-
-        """
+        """Start the kernel as soon as there is a connection."""
         self._process = time.time() + timeout
         self._timer.start()
 
     def startKernel(self):
-        """startKernel()
-
-        Launch the kernel in a subprocess, and connect to it via the
-        context and two Pypes.
-
-        """
+        """Launch the kernel in a subprocess and connect it via Yoton channels."""
 
         # Create channels
         self._create_channels()
 
         # Create info dict
-        info = {}
-        for key in self._info:
-            info[key] = self._info[key]
+        info = dict(self._info)
 
         # Send info stuff so that the kernel has access to the information
         self._stat_startup.send(info)
@@ -386,7 +373,7 @@ class KernelBroker:
         cwd = pyzo.pyzoDir
 
         # Host connection for the kernel to connect
-        # (tries several port numbers, staring from 'PYZO')
+        # (tries several port numbers, starting from 'PYZO')
         self._kernelCon = self._context.bind(
             "localhost:PYZO", max_tries=256, name="kernel"
         )
@@ -395,31 +382,37 @@ class KernelBroker:
         command = getCommandFromKernelInfo(self._info, self._kernelCon.port1)
         env = getEnvFromKernelInfo(self._info)
 
-        # Wrap command in call to 'cmd'?
-        if sys.platform.startswith("win"):
-            # as the author from Pype writes:
-            # if we don't run via a command shell, then either sometimes we
-            # don't get wx GUIs, or sometimes we can't kill the subprocesses.
-            # And I also see problems with Tk.
-            # But we only use it if we are sure that cmd is available.
-            # See pyzo issue #240
-            try:
-                subprocess.check_output('cmd /c "cd"', shell=True)
-            except (IOError, subprocess.SubprocessError):
-                pass  # Do not use cmd
-            else:
-                command = 'cmd /c "{}"'.format(command)
+        try:
+            # Wrap command in call to 'cmd'?
+            if sys.platform.startswith("win"):
+                # as the author from Pype writes:
+                # if we don't run via a command shell, then either sometimes we
+                # don't get wx GUIs, or sometimes we can't kill the subprocesses.
+                # And I also see problems with Tk.
+                # But we only use it if we are sure that cmd is available.
+                # See pyzo issue #240
+                try:
+                    subprocess.check_output('cmd /c "cd"', shell=True)
+                except (IOError, subprocess.SubprocessError):
+                    pass  # Do not use cmd
+                else:
+                    command = 'cmd /c "{}"'.format(command)
 
-        # Start process
-        self._process = subprocess.Popen(
-            command,
-            shell=True,
-            env=env,
-            cwd=cwd,
-            stdin=subprocess.PIPE,  # Fixes issue 165
-            stdout=subprocess.PIPE,
-            stderr=subprocess.STDOUT,
-        )
+            # Start process
+            self._process = subprocess.Popen(
+                command,
+                shell=True,
+                env=env,
+                cwd=cwd,
+                stdin=subprocess.PIPE,  # Fixes issue 165
+                stdout=subprocess.PIPE,
+                stderr=subprocess.STDOUT,
+            )
+        except Exception as err:
+            self._pending_restart = None
+            self._process = None
+            self._onKernelDied(str(err))
+            return
 
         # Set timeout for connection, i.e. after how much time of
         # unresponsive ness is the kernel found to be running extension code
@@ -443,11 +436,8 @@ class KernelBroker:
         self._pending_restart = None
 
     def hostConnectionForIDE(self, address="localhost"):
-        """hostConnectionForIDE()
-
-        Host a connection for an IDE to connect to. Returns the port to which
+        """Host a connection for an IDE to connect to. Returns the port to which
         the ide can connect.
-
         """
         c = self._context.bind(address + ":PYZO+256", max_tries=256, name="ide")
         return c.port1
@@ -455,11 +445,8 @@ class KernelBroker:
     ## Callbacks
 
     def _onKernelTimedOut(self, c, timedout):
-        """_onKernelTimedOut(c, timeout)
-
-        The kernel timed out (i.e. did not send heartbeat messages for
+        """The kernel timed out (i.e. did not send heartbeat messages for
         a while. It is probably running extension code.
-
         """
         if timedout:
             self._stat_interpreter.send("Very busy")
@@ -467,11 +454,7 @@ class KernelBroker:
             self._stat_interpreter.send("Busy")
 
     def _onKernelConnectionClose(self, c, why):
-        """_onKernelConnectionClose(c, why)
-
-        Connection with kernel lost. Tell clients why.
-
-        """
+        """Connection with kernel lost. Tell clients why."""
 
         # If we receive this event while the current kernel connection
         # is not the one that generated the event, ignore it.
@@ -485,16 +468,12 @@ class KernelBroker:
         if not self._terminator:
             self.terminate("because connecton was lost", "KILL", 0.5)
 
-    def _onKernelDied(self, returncode=0):
-        """_onKernelDied()
-
-        Kernel process died. Clean up!
-
-        """
+    def _onKernelDied(self, why=None):
+        """Kernel process died. Clean up!"""
 
         # If the kernel did not start yet, probably the command is invalid
         if self._kernelCon and self._kernelCon.is_waiting:
-            msg = "The process failed to start (invalid command?)."
+            msg = "The process failed to start."
         elif not self.isTerminating():
             msg = "Kernel process exited."
         elif not self._terminator._prev_action:
@@ -509,8 +488,11 @@ class KernelBroker:
 
         if self._context.connection_count:
             # Notify
-            returncodeMsg = "\n%s (%s)\n\n" % (msg, str(returncode))
-            self._strm_broker.send(returncodeMsg)
+            if isinstance(why, int):
+                total_msg = "\n{} ({})\n\n".format(msg, why)
+            else:
+                total_msg = "\n{}\n{}\n\n".format(msg, why)
+            self._strm_broker.send(total_msg)
             # Empty prompt and signal dead
             self._strm_prompt.send("\b")
             self._stat_interpreter.send("Dead")
@@ -526,11 +508,7 @@ class KernelBroker:
     ## Main loop and termination
 
     def terminate(self, reason="by user", action="TERM", timeout=0.0):
-        """terminate(reason='by user', action='TERM', timeout=0.0)
-
-        Initiate termination procedure for the current kernel.
-
-        """
+        """Initiate termination procedure for the current kernel."""
 
         # The terminatation procedure is started by creating
         # a KernelTerminator instance. This instance's iteration method
@@ -538,20 +516,14 @@ class KernelBroker:
         self._terminator = KernelTerminator(self, reason, action, timeout)
 
     def isTerminating(self):
-        """isTerminating()
-
-        Get whether the termination procedure has been initiated. This
-        simply checks whether there is a self._terminator instance.
+        """Get whether the termination procedure has been initiated.
 
+        This simply checks whether there is a self._terminator instance.
         """
         return bool(self._terminator)
 
     def mainLoopIter(self):
-        """mainLoopIter()
-
-        Periodically called. Kind of the main loop iteration for this kernel.
-
-        """
+        """Periodically called. Kind of the main loop iteration for this kernel."""
 
         # Get some important status info
         hasProcess = self._process is not None
@@ -592,6 +564,8 @@ class KernelBroker:
             for msg in self._ctrl_broker.recv_all():
                 if msg == "INT":
                     self._commandInterrupt()
+                elif msg == "PAUSE":
+                    self._commandPause()
                 elif msg == "TERM":
                     self._commandTerminate()
                 elif msg.startswith("RESTART"):
@@ -612,10 +586,19 @@ class KernelBroker:
             pid = self._kernelCon.pid2
             os.kill(pid, signal.SIGINT)
 
+    def _commandPause(self):
+        if self._process is None:
+            self._strm_broker.send("Cannot interrupt: process is dead.\n")
+        elif sys.platform == "win32":
+            self._reqp_introspect.interrupt(signum=signal.SIGFPE)
+        else:
+            pid = self._kernelCon.pid2
+            os.kill(pid, signal.SIGFPE)
+
     def _commandTerminate(self):
         # Start termination procedure
         # Kernel will receive term and act (if it can).
-        # If it wont, we will act in a second or so.
+        # If it won't, we will act in a second or so.
         if self._process is None:
             self._strm_broker.send("Cannot terminate: process is dead.\n")
         elif self.isTerminating():
@@ -647,8 +630,7 @@ class KernelBroker:
 
 
 class KernelTerminator:
-    """KernelTerminator(broker, reason='user terminated', action='TERM', timeout=0.0)
-
+    """
     Simple class to help terminating the kernel. It has a next() method
     that should be periodically called. It keeps track whether the timeout
     has passed and will undertake increaslingly ruder actions to terminate
@@ -728,8 +710,7 @@ class KernelTerminator:
 
 
 class StreamReader(threading.Thread):
-    """StreamReader(process, channel)
-
+    """
     Reads stdout of process and send to a yoton channel.
     This needs to be done in a separate thread because reading from
     a PYPE blocks.
@@ -737,12 +718,12 @@ class StreamReader(threading.Thread):
     """
 
     def __init__(self, process, strm_raw, strm_broker):
-        threading.Thread.__init__(self)
+        super().__init__()
 
         self._process = process
         self._strm_raw = strm_raw
         self._strm_broker = strm_broker
-        self.deamon = True
+        self.daemon = True
         self._exit = False
 
     def stop(self, timeout=1.0):
@@ -764,6 +745,8 @@ class StreamReader(threading.Thread):
             if not msg:  # or self._process.poll() is not None:
                 break
         # self._strm_broker.send('streamreader exit\n')
+        self._process.stdout.close()
+        self._process.stdin.close()
 
 
 class Kernelmanager:
@@ -787,17 +770,14 @@ class Kernelmanager:
         self._kernels = []
 
     def createKernel(self, info, name=None):
-        """create_kernel(info, name=None)
-
-        Create a new kernel. Returns the port number to connect to the
+        """Create a new kernel. Returns the port number to connect to the
         broker's context.
-
         """
 
         # Set name if not given
         if not name:
             i = len(self._kernels) + 1
-            name = "kernel %i" % i
+            name = "kernel {}".format(i)
 
         # Create kernel
         kernel = KernelBroker(self, info, name)
@@ -825,13 +805,10 @@ class Kernelmanager:
         return infos
 
     def terminateAll(self):
-        """terminateAll()
-
-        Terminates all kernels. Required when shutting down Pyzo.
+        """Terminates all kernels. Required when shutting down Pyzo.
         When this function returns, all kernels will be terminated.
-
         """
-        for kernel in [kernel for kernel in self._kernels]:
+        for kernel in self._kernels:
             # Try closing the process gently: by closing stdin
             terminator = KernelTerminator(kernel, "for closing down")
 
